@@ -19,6 +19,7 @@ from trytond import backend
 from trytond.pyson import Eval, Bool, If, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool
+from trytond.tools import firstline
 
 from trytond.ir.attachment import AttachmentCopyMixin
 from trytond.ir.note import NoteCopyMixin
@@ -73,15 +74,16 @@ class Purchase(
         }
     _depends = ['state']
 
-    company = fields.Many2One('company.company', 'Company', required=True,
+    company = fields.Many2One(
+        'company.company', "Company", required=True, select=True,
         states={
-            'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
+            'readonly': (
+                (Eval('state') != 'draft')
+                | Eval('lines', [0])
+                | Eval('party', True)
+                | Eval('invoice_party', True)),
             },
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
-        depends=['state'], select=True)
+        depends=['state'])
     number = fields.Char('Number', size=None, readonly=True, select=True)
     reference = fields.Char('Reference', select=True)
     description = fields.Char('Description', size=None, states=_states,
@@ -103,24 +105,33 @@ class Purchase(
             'readonly': ((Eval('state') != 'draft')
                 | (Eval('lines', [0]) & Eval('party'))),
             },
-        select=True, depends=['state'])
+        context={
+            'company': Eval('company', -1),
+            },
+        select=True, depends=['state', 'company'])
     party_lang = fields.Function(fields.Char('Party Language'),
         'on_change_with_party_lang')
     contact = fields.Many2One(
         'party.contact_mechanism', "Contact",
+        context={
+            'company': Eval('company', -1),
+            },
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['party'])
+        depends=['party', 'company'])
     invoice_party = fields.Many2One('party.party', "Invoice Party",
         states={
             'readonly': ((Eval('state') != 'draft')
                 | Eval('lines', [0])),
             },
+        context={
+            'company': Eval('company', -1),
+            },
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['state', 'party'])
+        depends=['state', 'party', 'company'])
     invoice_address = fields.Many2One('party.address', 'Invoice Address',
         domain=[
             ('party', '=', If(Bool(Eval('invoice_party')),
@@ -185,6 +196,11 @@ class Purchase(
     invoices_recreated = fields.Many2Many(
             'purchase.purchase-recreated-account.invoice',
             'purchase', 'invoice', 'Recreated Invoices', readonly=True)
+    origin = fields.Reference('Origin', selection='get_origin', select=True,
+        states={
+            'readonly': Eval('state') != 'draft',
+            },
+        depends=['state'])
     delivery_date = fields.Date(
         "Delivery Date",
         states={
@@ -227,8 +243,9 @@ class Purchase(
     @classmethod
     def __setup__(cls):
         super(Purchase, cls).__setup__()
+        cls.create_date.select = True
         cls._order = [
-            ('purchase_date', 'DESC'),
+            ('purchase_date', 'DESC NULLS FIRST'),
             ('id', 'DESC'),
             ]
         cls._transitions |= set((
@@ -266,7 +283,7 @@ class Purchase(
                             ()),
                         ],
                     'invisible': Eval('state') != 'draft',
-                    'readonly': ~Eval('lines', []),
+                    'readonly': ~Eval('lines', Eval('untaxed_amount', 0)),
                     'depends': ['state'],
                     },
                 'confirm': {
@@ -347,21 +364,11 @@ class Purchase(
                     values=['processing'],
                     where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
-        # Add index on create_date
-        table.index_action('create_date', action='add')
-
         # Migration from 5.6: rename state cancel to cancelled
         cursor.execute(*sql_table.update(
                 [sql_table.state], ['cancelled'],
                 where=sql_table.state == 'cancel'))
 
-    @classmethod
-    def default_payment_term(cls):
-        PaymentTerm = Pool().get('account.invoice.payment_term')
-        payment_terms = PaymentTerm.search(cls.payment_term.domain)
-        if len(payment_terms) == 1:
-            return payment_terms[0].id
-
     @classmethod
     def default_warehouse(cls):
         Location = Pool().get('stock.location')
@@ -371,6 +378,11 @@ class Purchase(
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('company')
+    def on_change_company(self):
+        self.invoice_method = self.default_invoice_method(
+            company=self.company.id if self.company else None)
+
     @staticmethod
     def default_state():
         return 'draft'
@@ -392,11 +404,12 @@ class Purchase(
             return company.currency.digits
         return 2
 
-    @staticmethod
-    def default_invoice_method():
+    @classmethod
+    def default_invoice_method(cls, **pattern):
         Configuration = Pool().get('purchase.configuration')
         configuration = Configuration(1)
-        return configuration.purchase_invoice_method
+        return configuration.get_multivalue(
+            'purchase_invoice_method', **pattern)
 
     @staticmethod
     def default_invoice_state():
@@ -408,39 +421,47 @@ class Purchase(
 
     @fields.depends('party', 'invoice_party', 'payment_term', 'lines')
     def on_change_party(self):
-        pool = Pool()
-        Currency = pool.get('currency.currency')
         cursor = Transaction().connection.cursor()
         table = self.__table__()
         if not self.invoice_party:
             self.invoice_address = None
-        self.payment_term = self.default_payment_term()
+        self.invoice_method = self.default_invoice_method()
         if not self.lines:
             self.currency = self.default_currency()
             self.currency_digits = self.default_currency_digits()
         if self.party:
             if not self.invoice_party:
                 self.invoice_address = self.party.address_get(type='invoice')
-            if self.party.supplier_payment_term:
-                self.payment_term = self.party.supplier_payment_term
 
             if not self.lines:
                 invoice_party = (
                     self.invoice_party.id if self.invoice_party else None)
-                subquery = table.select(table.currency,
+                subquery = table.select(
+                    table.currency,
+                    table.payment_term,
+                    table.invoice_method,
                     where=(table.party == self.party.id)
                     & (table.invoice_party == invoice_party),
                     order_by=table.id,
                     limit=10)
-                cursor.execute(*subquery.select(subquery.currency,
-                        group_by=subquery.currency,
+                cursor.execute(*subquery.select(
+                        subquery.currency,
+                        subquery.payment_term,
+                        subquery.invoice_method,
+                        group_by=[
+                            subquery.currency,
+                            subquery.payment_term,
+                            subquery.invoice_method,
+                            ],
                         order_by=Count(Literal(1)).desc))
                 row = cursor.fetchone()
                 if row:
-                    currency_id, = row
-                    self.currency = Currency(currency_id)
+                    self.currency, self.payment_term, self.invoice_method = row
                     self.currency_digits = self.currency.digits
 
+            if self.party.supplier_payment_term:
+                self.payment_term = self.party.supplier_payment_term
+
     @fields.depends('party', 'invoice_party')
     def on_change_invoice_party(self):
         if self.invoice_party:
@@ -463,11 +484,13 @@ class Purchase(
                 return self.party.lang.code
         return Config.get_language()
 
-    @fields.depends('party')
+    @fields.depends('party', 'company')
     def _get_tax_context(self):
         context = {}
         if self.party and self.party.lang:
             context['language'] = self.party.lang.code
+        if self.company:
+            context['company'] = self.company.id
         return context
 
     @fields.depends('lines', 'currency', methods=['get_tax_amount'])
@@ -475,7 +498,6 @@ class Purchase(
         self.untaxed_amount = Decimal('0.0')
         self.tax_amount = Decimal('0.0')
         self.total_amount = Decimal('0.0')
-        taxes = {}
         if self.lines:
             for line in self.lines:
                 self.untaxed_amount += getattr(line, 'amount', None) or 0
@@ -619,6 +641,18 @@ class Purchase(
                     'shipment_state': state,
                     })
 
+    @classmethod
+    def _get_origin(cls):
+        "Return list of Model names for origin Reference"
+        return ['purchase.purchase']
+
+    @classmethod
+    def get_origin(cls):
+        Model = Pool().get('ir.model')
+        get_name = Model.get_name
+        models = cls._get_origin()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
     @property
     def report_address(self):
         if self.invoice_address:
@@ -661,7 +695,7 @@ class Purchase(
 
     @classmethod
     def view_attributes(cls):
-        attributes = [
+        attributes = super().view_attributes() + [
             ('/form//field[@name="comment"]', 'spell', Eval('party_lang')),
             ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
             ('/tree/field[@name="invoice_state"]', 'visual',
@@ -719,14 +753,14 @@ class Purchase(
         Fill the number field with the purchase sequence
         '''
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('purchase.configuration')
 
         config = Config(1)
         for purchase in purchases:
             if purchase.number:
                 continue
-            purchase.number = Sequence.get_id(config.purchase_sequence.id)
+            purchase.number = config.get_multivalue(
+                'purchase_sequence', company=purchase.company.id).get()
         cls.save(purchases)
 
     @classmethod
@@ -1097,6 +1131,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
             'readonly': Eval('purchase_state') != 'draft',
             },
         depends=['purchase_state'])
+    summary = fields.Function(fields.Char('Summary'), 'on_change_with_summary')
     note = fields.Text('Note')
     taxes = fields.Many2Many('purchase.line-account.tax',
         'line', 'tax', 'Taxes',
@@ -1164,6 +1199,11 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         fields.Many2One('currency.currency', 'Currency'),
         'on_change_with_currency')
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('purchase')
+
     @classmethod
     def __register__(cls, module_name):
         super().__register__(module_name)
@@ -1357,9 +1397,9 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
     def on_change_unit(self):
         self.on_change_quantity()
 
-    @fields.depends(methods=['on_change_quantity'])
-    def on_change_taxes(self):
-        self.on_change_quantity()
+    @fields.depends('description')
+    def on_change_with_summary(self, name=None):
+        return firstline(self.description or '')
 
     @fields.depends('type', 'quantity', 'unit_price', 'unit', 'purchase',
         '_parent_purchase.currency')
@@ -1391,7 +1431,8 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
 
     def get_from_location(self, name):
         if (self.quantity or 0) >= 0:
-            return self.purchase.party.supplier_location.id
+            if self.purchase.party.supplier_location:
+                return self.purchase.party.supplier_location.id
         elif self.purchase.return_from_location:
             return self.purchase.return_from_location.id
 
@@ -1399,33 +1440,44 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         if (self.quantity or 0) >= 0:
             if self.purchase.warehouse:
                 return self.purchase.warehouse.input_location.id
-        else:
+        elif self.purchase.party.supplier_location:
             return self.purchase.party.supplier_location.id
 
-    @fields.depends('product_supplier', 'quantity', 'moves', 'purchase',
-        '_parent_purchase.purchase_date',
-        'delivery_date_edit', 'delivery_date_store',
-        '_parent_purchase.delivery_date')
+    @fields.depends('moves', methods=['planned_delivery_date'])
     def on_change_with_delivery_date(self, name=None):
-        pool = Pool()
-        Date = pool.get('ir.date')
         if self.moves:
             dates = filter(
                 None, (m.effective_date or m.planned_date for m in self.moves
                     if m.state != 'cancelled'))
             return min(dates, default=None)
+        return self.planned_delivery_date
+
+    @property
+    @fields.depends(
+        'product_supplier', 'quantity', 'purchase',
+        '_parent_purchase.purchase_date',
+        'delivery_date_edit', 'delivery_date_store',
+        '_parent_purchase.delivery_date', '_parent_purchase.party')
+    def planned_delivery_date(self):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        ProductSupplier = pool.get('purchase.product_supplier')
+        product_supplier = self.product_supplier
+        if not product_supplier:
+            product_supplier = ProductSupplier()
+            if self.purchase:
+                product_supplier.party = self.purchase.party
         delivery_date = None
         if self.delivery_date_edit:
             delivery_date = self.delivery_date_store
         elif self.purchase and self.purchase.delivery_date:
             delivery_date = self.purchase.delivery_date
-        elif (self.product_supplier
+        elif (product_supplier
                 and self.quantity is not None
                 and self.quantity > 0
                 and self.purchase):
             date = self.purchase.purchase_date if self.purchase else None
-            delivery_date = self.product_supplier.compute_supply_date(
-                date=date)
+            delivery_date = product_supplier.compute_supply_date(date=date)
             if delivery_date == datetime.date.max:
                 delivery_date = None
         if delivery_date and delivery_date < Date.today():
@@ -1462,6 +1514,8 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
 
         invoice_line = InvoiceLine()
         invoice_line.type = self.type
+        invoice_line.currency = self.currency
+        invoice_line.company = self.company
         invoice_line.description = self.description
         invoice_line.note = self.note
         invoice_line.origin = self
@@ -1485,9 +1539,12 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
         invoice_line.taxes = self.taxes
+        if self.company.purchase_taxes_expense:
+            invoice_line.taxes_deductible_rate = 0
+        elif self.product:
+            invoice_line.taxes_deductible_rate = (
+                self.product.supplier_taxes_deductible_rate_used)
         invoice_line.invoice_type = 'in'
-        invoice_line.currency = self.currency
-        invoice_line.company = self.company
         if self.product:
             invoice_line.account = self.product.account_expense_used
             if not invoice_line.account:
@@ -1498,7 +1555,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
                         product=self.product.rec_name))
         else:
             invoice_line.account = account_config.get_multivalue(
-                'default_category_account_expense')
+                'default_category_account_expense', company=self.company.id)
             if not invoice_line.account:
                 raise AccountError(
                     gettext('purchase'
@@ -1573,13 +1630,12 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         '''
         pool = Pool()
         Move = pool.get('stock.move')
-        Date = pool.get('ir.date')
 
         if self.type != 'line':
             return
         if not self.product:
             return
-        if self.product.type == 'service':
+        if self.product.type not in Move.get_product_types():
             return
 
         if (self.quantity >= 0) != (move_type == 'in'):
@@ -1608,15 +1664,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         if move.on_change_with_unit_price_required():
             move.unit_price = self.unit_price
             move.currency = self.purchase.currency
-        if self.moves:
-            # backorder can not be planned but delivery date could be used
-            # if set in the future
-            if self.delivery_date and self.delivery_date > Date.today():
-                move.planned_date = self.delivery_date
-            else:
-                move.planned_date = None
-        else:
-            move.planned_date = self.delivery_date
+        move.planned_date = self.planned_delivery_date
         move.invoice_lines = self._get_move_invoice_lines(move_type)
         move.origin = self
         return move
@@ -1689,7 +1737,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
 
     @classmethod
     def view_attributes(cls):
-        return [
+        return super().view_attributes() + [
             ('/form//field[@name="note"]|/form//field[@name="description"]',
                 'spell', Eval('_parent_purchase', {}).get('party_lang')),
             ('//label[@id="delivery_date"]', 'states', {
@@ -1717,6 +1765,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         default.setdefault('moves_ignored', None)
         default.setdefault('moves_recreated', None)
         default.setdefault('invoice_lines', None)
+        default.setdefault('actual_quantity')
         return super().copy(lines, default=default)
 
 
@@ -1760,10 +1809,10 @@ class PurchaseReport(CompanyReport):
             return super(PurchaseReport, cls).execute(ids, data)
 
     @classmethod
-    def get_context(cls, records, data):
+    def get_context(cls, records, header, data):
         pool = Pool()
         Date = pool.get('ir.date')
-        context = super(PurchaseReport, cls).get_context(records, data)
+        context = super().get_context(records, header, data)
         context['today'] = Date.today()
         return context
 
@@ -1784,7 +1833,7 @@ class OpenSupplier(Wizard):
 
         cursor.execute(*purchase.select(purchase.party,
                 group_by=purchase.party))
-        supplier_ids = [line[0] for line in cursor.fetchall()]
+        supplier_ids = [line[0] for line in cursor]
         action['pyson_domain'] = PYSONEncoder().encode(
             [('id', 'in', supplier_ids)])
         wizard = Wizard(ModelData.get_id('purchase', 'act_open_supplier'))
@@ -1986,6 +2035,7 @@ class ReturnPurchase(Wizard):
         purchases = self.records
         return_purchases = self.model.copy(purchases)
         for return_purchase, purchase in zip(return_purchases, purchases):
+            return_purchase.origin = purchase
             for line in return_purchase.lines:
                 if line.type == 'line':
                     line.quantity *= -1
